@@ -37,9 +37,10 @@ fn witness_keys_for_cert(cert_enum: &Certificate, keys: &mut BTreeSet<Ed25519Key
     }
 }
 
-fn min_fee(tx_builder: &TransactionBuilder) -> Result<Coin, JsError> {
-    let body = tx_builder.build()?;
-
+// tx_body must be the result of building from tx_builder
+// constructs the rest of the Transaction using fake witness data of the correct length
+// for use in calculating the size of the final Transaction
+fn fake_full_tx(tx_builder: &TransactionBuilder, body: TransactionBody) -> Result<Transaction, JsError> {
     let fake_key_root = Bip32PrivateKey::from_bip39_entropy(
         // art forum devote street sure rather head chuckle guard poverty release quote oak craft enemy
         &[0x0c, 0xcb, 0x74, 0xf3, 0x6b, 0x7d, 0xa1, 0x64, 0x9a, 0x81, 0x44, 0x67, 0x55, 0x22, 0xd4, 0xd8, 0x09, 0x7c, 0x64, 0x12],
@@ -92,11 +93,15 @@ fn min_fee(tx_builder: &TransactionBuilder) -> Result<Coin, JsError> {
         plutus_data: None,
         redeemers: None,
     };
-    let full_tx = Transaction {
+    Ok(Transaction {
         body,
         witness_set,
         auxiliary_data: tx_builder.auxiliary_data.clone(),
-    };
+    })
+}
+
+fn min_fee(tx_builder: &TransactionBuilder) -> Result<Coin, JsError> {
+    let full_tx = fake_full_tx(tx_builder, tx_builder.build()?)?;
     fees::min_fee(&full_tx, &tx_builder.fee_algo)
 }
 
@@ -121,6 +126,8 @@ pub struct TransactionBuilder {
     minimum_utxo_val: BigNum,
     pool_deposit: BigNum,
     key_deposit: BigNum,
+    max_output_size: u32,
+    max_tx_size: u32,
     fee_algo: fees::LinearFee,
     inputs: Vec<TxBuilderInput>,
     outputs: TransactionOutputs,
@@ -214,7 +221,7 @@ impl TransactionBuilder {
         let mut self_copy = self.clone();
 
         // we need some value for these for it to be a a valid transaction
-        // but since we're only calculating the different between the fee of two transactions
+        // but since we're only calculating the difference between the fee of two transactions
         // it doesn't matter what these are set as, since it cancels out
         self_copy.set_fee(&to_bignum(0));
 
@@ -226,6 +233,14 @@ impl TransactionBuilder {
     }
 
     pub fn add_output(&mut self, output: &TransactionOutput) -> Result<(), JsError> {
+        let output_size = output.to_bytes().len();
+        if output_size > self.max_output_size as usize {
+            return Err(JsError::from_str(&format!(
+                "Maximum output size of {} exceeded. Found: {}",
+                self.max_output_size,
+                output_size
+            )));
+        }
         let min_ada = min_ada_required(&output.amount(), &self.minimum_utxo_val);
         if output.amount().coin() < min_ada {
             Err(JsError::from_str(&format!(
@@ -240,7 +255,7 @@ impl TransactionBuilder {
     }
 
     /// calculates how much the fee would increase if you added a given output
-    pub fn fee_for_output(&mut self, output: &TransactionOutput) -> Result<Coin, JsError> {
+    pub fn fee_for_output(&self, output: &TransactionOutput) -> Result<Coin, JsError> {
         let mut self_copy = self.clone();
 
         // we need some value for these for it to be a a valid transaction
@@ -291,11 +306,15 @@ impl TransactionBuilder {
         minimum_utxo_val: &Coin,
         pool_deposit: &BigNum, // protocol parameter
         key_deposit: &BigNum,  // protocol parameter
+        max_output_size: u32, // protocol parameter
+        max_tx_size: u32, // protocol parameter
     ) -> Self {
         Self {
             minimum_utxo_val: minimum_utxo_val.clone(),
             key_deposit: key_deposit.clone(),
             pool_deposit: pool_deposit.clone(),
+            max_output_size,
+            max_tx_size,
             fee_algo: linear_fee.clone(),
             inputs: Vec::new(),
             outputs: TransactionOutputs::new(),
@@ -383,19 +402,97 @@ impl TransactionBuilder {
             },
             Some(Ordering::Less) => Err(JsError::from_str("Insufficient input in transaction")),
             Some(Ordering::Greater) => {
+                fn has_assets(ma: Option<MultiAsset>) -> bool {
+                    ma.map(|assets| assets.len() > 0).unwrap_or(false)
+                }
                 let change_estimator = input_total.checked_sub(&output_total)?;
-                let min_ada = min_ada_required(&change_estimator.clone(), &self.minimum_utxo_val);
+                if has_assets(change_estimator.multiasset()) {
+                    fn pack_nfts_for_change(max_output_size: u32, change_address: &Address, change_estimator: &Value) -> Result<MultiAsset, JsError> {
+                        // we insert the entire available ADA temporarily here since that could potentially impact the size
+                        // as it could be 1, 2 3 or 4 bytes for Coin.
+                        let mut base_coin = Value::new(&change_estimator.coin());
+                        base_coin.set_multiasset(&MultiAsset::new());
+                        let mut output = TransactionOutput::new(change_address, &base_coin);
+                        // If this becomes slow on large TXs we can optimize it like the folowing
+                        // to avoid cloning + reserializing the entire output.
+                        // This would probably be more relevant if we use a smarter packing algorithm
+                        // which might need to compare more size differences than greedy
+                        //let mut bytes_used = output.to_bytes().len();
 
-                let enough_input = |
-                    builder: &mut TransactionBuilder,
-                    burn_fee: &dyn Fn(&mut TransactionBuilder, &BigNum) -> Result<bool, JsError>,
-                    add_change: &dyn Fn(&mut TransactionBuilder, &BigNum) -> Result<bool, JsError>
-                | {
+                        // a greedy packing is done here to avoid an exponential bin-packing
+                        // which in most cases likely shouldn't be the difference between
+                        // having an extra change output or not unless there are gigantic
+                        // differences in NFT policy sizes
+                        for (policy, assets) in change_estimator.multiasset().unwrap().0.iter() {
+                            // for simplicity we also don't split assets within a single policy since
+                            // you would need to have a very high amoun of assets (which add 1-36 bytes each)
+                            // in a single policy to make a difference. In the future if this becomes an issue
+                            // we can change that here.
+
+                            // this is the other part of the optimization but we need to take into account
+                            // the difference between CBOR encoding which can change which happens in two places:
+                            // a) length within assets of one policy id
+                            // b) length of the entire multiasset
+                            // so for simplicity we will just do it the safe, naive way unless
+                            // performance becomes an issue.
+                            //let extra_bytes = policy.to_bytes().len() + assets.to_bytes().len() + 2 + cbor_len_diff;
+                            //if bytes_used + extra_bytes <= max_output_size as usize {
+                            let old_amount = output.amount.clone();
+                            let mut val = Value::new(&Coin::zero());
+                            let mut next_nft = MultiAsset::new();
+                            next_nft.insert(policy, assets);
+                            val.set_multiasset(&next_nft);
+                            output.amount = output.amount.checked_add(&val)?;
+                            if output.to_bytes().len() > max_output_size as usize {
+                                output.amount = old_amount;
+                                break;
+                            }
+                        }
+                        Ok(output.amount.multiasset().unwrap())
+                    }
+                    let mut change_left = input_total.checked_sub(&output_total)?;
+                    let mut new_fee = fee.clone();
+                    // we might need multiple change outputs for cases where the change has many asset types
+                    // which surpass the max UTXO size limit
+                    while let Some(Ordering::Greater) = change_left.multiasset.as_ref().map_or_else(|| None, |ma| ma.partial_cmp(&MultiAsset::new())) {
+                        let nft_change = pack_nfts_for_change(self.max_output_size, address, &change_left)?;
+                        if nft_change.len() == 0 {
+                            // this likely should never happen
+                            return Err(JsError::from_str("NFTs too large for change output"));
+                        }
+                        // we only add the minimum needed (for now) to cover this output
+                        let mut change_value = Value::new(&Coin::zero());
+                        change_value.set_multiasset(&nft_change);
+                        let min_ada = min_ada_required(&change_value, &self.minimum_utxo_val);
+                        change_value.set_coin(&min_ada);
+                        let change_output = TransactionOutput::new(address, &change_value);
+                        // increase fee
+                        let fee_for_change = self.fee_for_output(&change_output)?;
+                        new_fee = new_fee.checked_add(&fee_for_change)?;
+                        if change_left.coin() < min_ada.checked_add(&new_fee)? {
+                            return Err(JsError::from_str("Not enough ADA leftover to include non-ADA assets in a change address"));
+                        }
+                        change_left = change_left.checked_sub(&change_value)?;
+                        self.add_output(&change_output)?;
+                    }
+                    change_left = change_left.checked_sub(&Value::new(&new_fee))?;
+                    self.set_fee(&new_fee);
+                    // add in the rest of the ADA
+                    self.outputs.0.last_mut().unwrap().amount = self.outputs.0.last().unwrap().amount.checked_add(&change_left)?;
+                    Ok(true)
+                } else {
+                    let min_ada = min_ada_required(&change_estimator, &self.minimum_utxo_val);
+                    // no-asset case so we have no problem burning the rest if there is no other option
+                    fn burn_extra(builder: &mut TransactionBuilder, burn_amount: &BigNum) -> Result<bool, JsError> {
+                        // recall: min_fee assumed the fee was the maximum possible so we definitely have enough input to cover whatever fee it ends up being
+                        builder.set_fee(burn_amount);
+                        Ok(false) // not enough input to covert the extra fee from adding an output so we just burn whatever is left
+                    };
                     match change_estimator.coin() >= min_ada {
-                        false => burn_fee(builder, &change_estimator.coin()),
+                        false => burn_extra(self, &change_estimator.coin()),
                         true => {
                             // check how much the fee would increase if we added a change output
-                            let fee_for_change = builder.fee_for_output(&TransactionOutput {
+                            let fee_for_change = self.fee_for_output(&TransactionOutput {
                                 address: address.clone(),
                                 amount: change_estimator.clone(),
                                 // TODO: data hash?
@@ -404,57 +501,31 @@ impl TransactionBuilder {
 
                             let new_fee = fee.checked_add(&fee_for_change)?;
                             match change_estimator.coin() >= min_ada.checked_add(&Value::new(&new_fee).coin())? {
-                                false => burn_fee(builder, &change_estimator.coin()),
-                                true => add_change(builder, &new_fee)
+                                false => burn_extra(self, &change_estimator.coin()),
+                                true => {
+                                    // recall: min_fee assumed the fee was the maximum possible so we definitely have enough input to cover whatever fee it ends up being
+                                    self.set_fee(&new_fee);
+
+                                    self.add_output(&TransactionOutput {
+                                        address: address.clone(),
+                                        amount: change_estimator.checked_sub(&Value::new(&new_fee.clone()))?,
+                                        data_hash: None, // TODO: How do we get DataHash?
+                                    })?;
+
+                                    Ok(true)
+                                }
                             }
                         }
                     }
-                };
-
-                let burn_extra = |
-                    builder: &mut TransactionBuilder,
-                    burn_amount: &BigNum
-                | {
-                    let has_assets = change_estimator.multiasset().map(|assets| assets.len() > 0).unwrap_or(false);
-                    match has_assets {
-                        false => {
-                            // recall: min_fee assumed the fee was the maximum possible so we definitely have enough input to cover whatever fee it ends up being
-                            builder.set_fee(burn_amount);
-                            Ok(false) // not enough input to covert the extra fee from adding an output so we just burn whatever is left
-                        },
-                        true => Err(JsError::from_str("Not enough ADA leftover to include non-ADA assets in a change address")),
-                    }
-                };
-
-                let add_change = |
-                    builder: &mut TransactionBuilder,
-                    new_fee: &BigNum
-                | {
-                    // recall: min_fee assumed the fee was the maximum possible so we definitely have enough input to cover whatever fee it ends up being
-                    builder.set_fee(&new_fee);
-
-                    builder.add_output(&TransactionOutput {
-                        address: address.clone(),
-                        amount: change_estimator.checked_sub(&Value::new(&new_fee.clone()))?,
-                        data_hash: None, // TODO: How do we get DataHash?
-                    })?;
-
-                    Ok(true)
-                };
-
-                enough_input(
-                    self,
-                    &burn_extra,
-                    &add_change
-                )
+                }
             }
             None => Err(JsError::from_str("missing input for some native asset")),
         }
     }
 
-    pub fn build(&self) -> Result<TransactionBody, JsError> {
+    fn build_and_size(&self) -> Result<(TransactionBody, usize), JsError> {
         let fee = self.fee.ok_or_else(|| JsError::from_str("Fee not specified"))?;
-        Ok(TransactionBody {
+        let built = TransactionBody {
             inputs: TransactionInputs(self.inputs.iter().map(|ref tx_builder_input| tx_builder_input.input.clone()).collect()),
             outputs: self.outputs.clone(),
             fee: fee,
@@ -473,7 +544,32 @@ impl TransactionBuilder {
             collateral: None,
             required_signers: None,
             network_id: None,
-        })
+        };
+        // we must build a tx with fake data (of correct size) to check the final Transaction size
+        let full_tx = fake_full_tx(self, built)?;
+        let full_tx_size = full_tx.to_bytes().len();
+        return Ok((full_tx.body, full_tx_size));
+    }
+
+    pub fn full_size(&self) -> Result<usize, JsError> {
+        return self.build_and_size().map(|r| { r.1 });
+    }
+
+    pub fn output_sizes(&self) -> Vec<usize> {
+        return self.outputs.0.iter().map(|o| { o.to_bytes().len() }).collect();
+    }
+
+    pub fn build(&self) -> Result<TransactionBody, JsError> {
+        let (body, full_tx_size) = self.build_and_size()?;
+        if full_tx_size > self.max_tx_size as usize {
+            Err(JsError::from_str(&format!(
+                "Maximum transaction size of {} exceeded. Found: {}",
+                self.max_tx_size,
+                full_tx_size
+            )))
+        } else {
+            Ok(body)
+        }
     }
 
     /// warning: sum of all parts of a transaction must equal 0. You cannot just set the fee to the min value and forget about it
@@ -491,6 +587,9 @@ mod tests {
     use super::*;
     use fees::*;
 
+    const MAX_OUTPUT_SIZE: u32 = 4000;
+    const MAX_TX_SIZE: u32 = 8000; // might be out of date but suffices for our tests
+
     fn genesis_id() -> TransactionHash {
         TransactionHash::from([0u8; TransactionHash::BYTE_COUNT])
     }
@@ -508,8 +607,14 @@ mod tests {
     #[test]
     fn build_tx_with_change() {
         let linear_fee = LinearFee::new(&to_bignum(500), &to_bignum(2));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(1), &to_bignum(1));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(1),
+            &to_bignum(1),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -557,14 +662,22 @@ mod tests {
             tx_builder.get_explicit_input().unwrap().checked_add(&tx_builder.get_implicit_input().unwrap()).unwrap(),
             tx_builder.get_explicit_output().unwrap().checked_add(&Value::new(&tx_builder.get_fee_if_set().unwrap())).unwrap()
         );
+        assert_eq!(tx_builder.full_size().unwrap(), 283);
+        assert_eq!(tx_builder.output_sizes(), vec![61, 65]);
         let _final_tx = tx_builder.build(); // just test that it doesn't throw
     }
 
     #[test]
     fn build_tx_without_change() {
         let linear_fee = LinearFee::new(&to_bignum(500), &to_bignum(2));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(1), &to_bignum(1));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(1),
+            &to_bignum(1),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -623,6 +736,8 @@ mod tests {
             &to_bignum(1),
             &to_bignum(1),
             &to_bignum(1_000_000),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
         );
         let spend = root_key_15()
             .derive(harden(1852))
@@ -685,8 +800,14 @@ mod tests {
     fn build_tx_exact_amount() {
         // transactions where sum(input) == sum(output) exact should pass
         let linear_fee = LinearFee::new(&to_bignum(0), &to_bignum(0));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(0), &to_bignum(0));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(0),
+            &to_bignum(0),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -736,8 +857,14 @@ mod tests {
     fn build_tx_exact_change() {
         // transactions where we have exactly enough ADA to add change should pass
         let linear_fee = LinearFee::new(&to_bignum(0), &to_bignum(0));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(0), &to_bignum(0));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(0),
+            &to_bignum(0),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -796,8 +923,14 @@ mod tests {
     fn build_tx_insufficient_deposit() {
         // transactions should fail with insufficient fees if a deposit is required
         let linear_fee = LinearFee::new(&to_bignum(0), &to_bignum(0));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(0), &to_bignum(5));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(0),
+            &to_bignum(5),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -861,8 +994,14 @@ mod tests {
     #[test]
     fn build_tx_with_inputs() {
         let linear_fee = LinearFee::new(&to_bignum(500), &to_bignum(2));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(1), &to_bignum(1));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(1),
+            &to_bignum(1),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -941,6 +1080,8 @@ mod tests {
             &minimum_utxo_value,
             &to_bignum(0),
             &to_bignum(0),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
         );
         let spend = root_key_15()
             .derive(harden(1852))
@@ -1050,8 +1191,14 @@ mod tests {
     #[should_panic]
     fn build_tx_leftover_assets() {
         let linear_fee = LinearFee::new(&to_bignum(500), &to_bignum(2));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1), &to_bignum(1), &to_bignum(1));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1),
+            &to_bignum(1),
+            &to_bignum(1),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
         let spend = root_key_15()
             .derive(harden(1852))
             .derive(harden(1815))
@@ -1118,8 +1265,14 @@ mod tests {
     #[test]
     fn build_tx_burn_less_than_min_ada() {
         let linear_fee = LinearFee::new(&to_bignum(44), &to_bignum(155381));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1000000), &to_bignum(500000000), &to_bignum(2000000));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1000000),
+            &to_bignum(500000000),
+            &to_bignum(2000000),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
 
         let output_addr = ByronAddress::from_base58("Ae2tdPwUPEZD9QQf2ZrcYV34pYJwxK4vqXaF8EXkup1eYH73zUScHReM42b").unwrap();
         tx_builder.add_output(&TransactionOutput::new(
@@ -1154,8 +1307,14 @@ mod tests {
     #[test]
     fn build_tx_burn_empty_assets() {
         let linear_fee = LinearFee::new(&to_bignum(44), &to_bignum(155381));
-        let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1000000), &to_bignum(500000000), &to_bignum(2000000));
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &to_bignum(1000000),
+            &to_bignum(500000000),
+            &to_bignum(2000000),
+            MAX_OUTPUT_SIZE,
+            MAX_TX_SIZE
+        );
 
         let output_addr = ByronAddress::from_base58("Ae2tdPwUPEZD9QQf2ZrcYV34pYJwxK4vqXaF8EXkup1eYH73zUScHReM42b").unwrap();
         tx_builder.add_output(&TransactionOutput::new(
@@ -1193,7 +1352,14 @@ mod tests {
     fn build_tx_no_useless_multiasset() {
         let linear_fee = LinearFee::new(&to_bignum(44), &to_bignum(155381));
         let mut tx_builder =
-            TransactionBuilder::new(&linear_fee, &to_bignum(1000000), &to_bignum(500000000), &to_bignum(2000000));
+            TransactionBuilder::new(
+                &linear_fee,
+                &to_bignum(1000000),
+                &to_bignum(500000000),
+                &to_bignum(2000000),
+                MAX_OUTPUT_SIZE,
+                MAX_TX_SIZE,
+            );
 
         let policy_id = &PolicyID::from([0u8; 28]);
         let name = AssetName::new(vec![0u8, 1, 2, 3]).unwrap();
@@ -1232,7 +1398,7 @@ mod tests {
             &output_addr.to_address(),
             &output_amount
         )).unwrap();
-        
+
         tx_builder.set_ttl(1);
 
         let change_addr = ByronAddress::from_base58("Ae2tdPwUPEZGUEsuMAhvDcy94LKsZxDjCbgaiBBMgYpR8sKf96xJmit7Eho").unwrap();
@@ -1249,4 +1415,172 @@ mod tests {
         // the change should be only ADA and not have any multiasset struct in it
         assert!(change_assets.is_none());
     }
+
+    #[test]
+    fn build_tx_add_change_split_nfts() {
+        let linear_fee = LinearFee::new(&to_bignum(0), &to_bignum(1));
+        let minimum_utxo_value = to_bignum(1);
+        let max_output_size = 150; // super low max output size to test with fewer assets
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &minimum_utxo_value,
+            &to_bignum(0),
+            &to_bignum(0),
+            max_output_size,
+            MAX_TX_SIZE
+        );
+
+        let policy_ids = [
+            PolicyID::from([0u8; 28]),
+            PolicyID::from([1u8; 28]),
+            PolicyID::from([2u8; 28]),
+        ];
+        let names = [
+            AssetName::new(vec![99u8; 32]).unwrap(),
+            AssetName::new(vec![0u8, 1, 2, 3]).unwrap(),
+            AssetName::new(vec![4u8, 5, 6, 7, 8, 9]).unwrap(),
+        ];
+
+        let multiasset = policy_ids
+            .iter()
+            .zip(names.iter())
+            .fold(MultiAsset::new(), |mut acc, (policy_id, name)| {
+                acc.insert(policy_id, &{
+                    let mut assets = Assets::new();
+                    assets.insert(&name, &to_bignum(500));
+                    assets
+                });
+                acc
+            });
+
+        let mut input_value = Value::new(&to_bignum(10));
+        input_value.set_multiasset(&multiasset);
+
+        tx_builder.add_input(
+            &ByronAddress::from_base58("Ae2tdPwUPEZ5uzkzh1o2DHECiUi3iugvnnKHRisPgRRP3CTF4KCMvy54Xd3").unwrap().to_address(),
+            &TransactionInput::new(
+                &genesis_id(),
+                0
+            ),
+            &input_value
+        );
+
+        let output_addr = ByronAddress::from_base58("Ae2tdPwUPEZD9QQf2ZrcYV34pYJwxK4vqXaF8EXkup1eYH73zUScHReM42b").unwrap().to_address();
+        let output_amount = Value::new(&to_bignum(1));
+
+        tx_builder
+            .add_output(&TransactionOutput::new(&output_addr, &output_amount))
+            .unwrap();
+
+        let change_addr = ByronAddress::from_base58("Ae2tdPwUPEZGUEsuMAhvDcy94LKsZxDjCbgaiBBMgYpR8sKf96xJmit7Eho").unwrap().to_address();
+
+        let added_change = tx_builder.add_change_if_needed(&change_addr).unwrap();
+        assert_eq!(added_change, true);
+        let final_tx = tx_builder.build().unwrap();
+        assert_eq!(final_tx.outputs().len(), 3);
+        for (policy_id, asset_name) in policy_ids.iter().zip(names.iter()) {
+            assert!(final_tx
+                .outputs
+                .0
+                .iter()
+                .find(|output| output.amount.multiasset.as_ref().map_or_else(
+                    || false,
+                    |ma| ma.0.iter().find(|(pid, a)| *pid == policy_id
+                        && a.0.iter().find(|(name, _)| *name == asset_name).is_some()).is_some()
+                )).is_some()
+            );
+        }
+        for output in final_tx.outputs.0.iter() {
+            assert!(output.to_bytes().len() <= max_output_size as usize);
+        }
+    }
+
+    #[test]
+    fn build_tx_too_big_output() {
+        let linear_fee = LinearFee::new(&to_bignum(0), &to_bignum(1));
+        let minimum_utxo_value = to_bignum(1);
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &minimum_utxo_value,
+            &to_bignum(0),
+            &to_bignum(0),
+            10, // super low max output size to test,
+            MAX_TX_SIZE
+        );
+
+        tx_builder.add_input(
+            &ByronAddress::from_base58("Ae2tdPwUPEZ5uzkzh1o2DHECiUi3iugvnnKHRisPgRRP3CTF4KCMvy54Xd3").unwrap().to_address(),
+            &TransactionInput::new(
+                &genesis_id(),
+                0
+            ),
+            &Value::new(&to_bignum(10))
+        );
+
+        let output_addr = ByronAddress::from_base58("Ae2tdPwUPEZD9QQf2ZrcYV34pYJwxK4vqXaF8EXkup1eYH73zUScHReM42b").unwrap().to_address();
+        let output_amount = Value::new(&to_bignum(1));
+
+        assert!(tx_builder.add_output(&TransactionOutput::new(&output_addr, &output_amount)).is_err());
+    }
+
+    #[test]
+    fn build_tx_add_change_nfts_not_enough_ada() {
+        let linear_fee = LinearFee::new(&to_bignum(0), &to_bignum(1));
+        let minimum_utxo_value = to_bignum(1);
+        let max_output_size = 150; // super low max output size to test with fewer assets
+        let mut tx_builder = TransactionBuilder::new(
+            &linear_fee,
+            &minimum_utxo_value,
+            &to_bignum(0),
+            &to_bignum(0),
+            max_output_size,
+            MAX_TX_SIZE
+        );
+
+        let policy_ids = [
+            PolicyID::from([0u8; 28]),
+            PolicyID::from([1u8; 28]),
+            PolicyID::from([2u8; 28]),
+        ];
+        let names = [
+            AssetName::new(vec![99u8; 32]).unwrap(),
+            AssetName::new(vec![0u8, 1, 2, 3]).unwrap(),
+            AssetName::new(vec![4u8, 5, 6, 7, 8, 9]).unwrap(),
+        ];
+
+        let multiasset = policy_ids
+            .iter()
+            .zip(names.iter())
+            .fold(MultiAsset::new(), |mut acc, (policy_id, name)| {
+                acc.insert(policy_id, &{
+                    let mut assets = Assets::new();
+                    assets.insert(&name, &to_bignum(500));
+                    assets
+                });
+                acc
+            });
+
+        let mut input_value = Value::new(&to_bignum(2));
+        input_value.set_multiasset(&multiasset);
+
+        tx_builder.add_input(
+            &ByronAddress::from_base58("Ae2tdPwUPEZ5uzkzh1o2DHECiUi3iugvnnKHRisPgRRP3CTF4KCMvy54Xd3").unwrap().to_address(),
+            &TransactionInput::new(
+                &genesis_id(),
+                0
+            ),
+            &input_value
+        );
+
+        let output_addr = ByronAddress::from_base58("Ae2tdPwUPEZD9QQf2ZrcYV34pYJwxK4vqXaF8EXkup1eYH73zUScHReM42b").unwrap().to_address();
+        let output_amount = Value::new(&to_bignum(1));
+
+        tx_builder
+            .add_output(&TransactionOutput::new(&output_addr, &output_amount))
+            .unwrap();
+
+        let change_addr = ByronAddress::from_base58("Ae2tdPwUPEZGUEsuMAhvDcy94LKsZxDjCbgaiBBMgYpR8sKf96xJmit7Eho").unwrap().to_address();
+
+        assert!(tx_builder.add_change_if_needed(&change_addr).is_err())
+    }
 }
